__author__ = "Arijit Basu"
__description__ = "Viper is a handy tool for easily running infrastructure management tasks and commands."
__email__ = "sayanarijit@gmail.com"
__homepage__ = "https://github.com/sayanarijit/viper"
__license__ = "MIT"
__version__ = "v0.0.1"

from viper.collections import (
    Host,
    Hosts,
    Task,
    TaskResult,
    TaskResults,
    TaskRunner,
    TaskRunners,
)
from viper.demo import __doc__ as __viper_cli_examples__

__doc__ = f"""{__description__}

{__viper_cli_examples__}
"""
